﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Xml;
using System.Xml.Linq;
using System.Xml.XPath;
using XmlClassifier.Algorithms.DistanceMeasures;
using XmlClassifier.Helpers;

namespace XmlClassifier.DataStructures.Patterns
{
    /// <summary>
    /// 
    /// </summary>
    public class PathPattern : BasePattern
    {
        private static int maximalPathLength = int.MaxValue;
        private static bool maximalPathsOnly = false;

        public static int MaximalPatternLength
        {
            get
            {
                return maximalPathLength;
            }
            set
            {
                if (value == 0)
                {
                    maximalPathLength = int.MaxValue;
                }
                else
                {
                    maximalPathLength = value;
                }
            }
        }
        public static bool MaximalPathsOnly
        {
            get
            {
                return maximalPathsOnly;
            }
            set
            {
                maximalPathsOnly = value;
            }
        }

        public string Pattern { get; private set; }

		private double? size;

		public override double Size
		{
			get
			{
				if (size.HasValue)
				{
					return size.Value;
				}
				else
				{
					return this.Pattern.Length;
				}
			}
			set
			{
				this.size = value;
			}
		}

        /// <summary>
        /// 
        /// </summary>
        /// <param name="document"></param>
        /// <returns></returns>
        public override bool Matches(Document document)
        {
            var res = document.XDocument.XPathSelectElements("/" + this.Pattern);
            return res != null && res.Count() > 0;
        }

        public override int CountMatches(Document document)
        {
            var res = document.XDocument.XPathSelectElements("/" + this.Pattern);
            return res == null ? 0 : res.Count();
        }

        public override List<BasePattern> GetPatterns(List<Document> documents, double minsup)
        {
            List<PathPattern> result = new List<PathPattern>();

            int numberOfClasses = documents.Select(x => x.ExpectedClass).Distinct().Count();
            for (int i = 0; i < numberOfClasses; i++)
            {
                var tmpDocuments = documents.Where(x => x.ExpectedClass == i).ToList();
                var tmp = this.GetPatternsInSingleClass(tmpDocuments, minsup);
                tmp.ForEach(x => { x.Class = i; });

                result.AddRange(tmp);
            }

            return result.ToList<BasePattern>();
        }

        public static void TransformDocuments(List<Document> documents)
        {
            foreach (var document in documents)
            {
                string[] t = document.Content.Split(' ');

                XElement curr = new XElement("e" + t[3]);
                XElement root = new XElement("Root", curr);
                for (int i = 4; i < t.Length; i++)
                {
                    if (int.Parse(t[i]) >= 0)
                    {
                        var tmp = new XElement("e" + t[i]);
                        curr.Add(tmp);
                        curr = tmp;
                    }
                    else
                    {
                        curr = curr.Parent;
                    }
                }

                document.XDocument = root;
            }
        }

        private List<PathPattern> GetPatternsInSingleClass(List<Document> documents, double minsup)
        {
            List<XElement> xdocuments = documents.Select(x => x.XDocument).ToList();

            int nodmin = (int)Math.Round(minsup * xdocuments.Count);

            List<PathPattern> result = new List<PathPattern>();
            List<PathPattern> candidates = GetFirstCandidates(xdocuments);

            int patternLength = 1;
            while (candidates.Count > 0 && patternLength <= MaximalPatternLength)
            {
				DebugConsole.WriteLine("Permutacje " + patternLength + "-elementowe: " + candidates.Count + " kandydatów.");

                int i = candidates.Count - 1;
                while (i >= 0)
                {
					DebugConsole.SetCursorPosition(0, Console.CursorTop);
					DebugConsole.Write(string.Format("{0,3:0}%", (double)(candidates.Count - i) * 100 / candidates.Count));

                    foreach (var d in xdocuments)
                    {
                        if (d.XPathSelectElement("/" + candidates[i].Pattern) != null)
                        {
                            candidates[i].Support++;
                        }
                    }

                    if (candidates[i].Support >= nodmin)
                    {
                        result.Add(new PathPattern() { Pattern = candidates[i].Pattern, Support = candidates[i].Support });
                        if (MaximalPathsOnly)
                        {
                            RemovePartialPaths(result, candidates[i].Pattern);
                        }
                    }
                    else
                    {
                        candidates.RemoveAt(i);
                    }

                    i--;
                }

				DebugConsole.SetCursorPosition(0, Console.CursorTop);

                candidates = Permutate(candidates);
                patternLength++;
            }

            return result;
        }

        /// <summary>
        /// Usuwa dwie podścieżki, z których została złożona przekazana ścieżka
        /// </summary>
        /// <param name="result">Zbiór, w którego mają zostać usunięte ścieżki</param>
        /// <param name="path">Ścieżka</param>
        private static void RemovePartialPaths(List<PathPattern> result, string path)
        {
            if (path.Count(x => x == '/') > 1)
            {
                string left = path.Substring(0, path.LastIndexOf('/'));
                string right = path.Substring(path.IndexOf('/', path.IndexOf('/') + 1));

                result.RemoveAll(x => x.Pattern == right || x.Pattern == left);
            }
        }

        private static List<PathPattern> GetFirstCandidates(List<XElement> documents)
        {
            List<PathPattern> result = new List<PathPattern>();

            foreach (var d in documents)
            {
                foreach (var v in d.XPathSelectElements("//*"))
                {
                    if (!result.Any(x => x.Pattern == "/" + v.Name.LocalName))
                    {
                        result.Add(new PathPattern() { Pattern = "/" + v.Name.LocalName, Support = 0 });
                    }
                }
            }

            return result;
        }

        private static List<PathPattern> Permutate(List<PathPattern> prev)
        {
            List<PathPattern> result = new List<PathPattern>();

            foreach (var p in prev)
            {
                string prefix = p.Pattern.Substring(0, p.Pattern.LastIndexOf('/'));
                foreach (var n in prev.Where(x => x.Pattern.EndsWith(prefix)))
                {
                    result.Add(new PathPattern() { Pattern = n.Pattern + p.Pattern.Substring(p.Pattern.LastIndexOf('/')), Support = 0 });
                }
            }

            return result;
        }

        public static int MaximalPathLength { get; set; }

        public override double CalculateDistance(Document document)
        {
            List<string> paths = GetFullPaths(document);

            // dla każdej ścieżki policz lewensztajna
            double dist = 1;
            double min = 1;
            foreach (var path in paths)
            {
                dist = PED.CaluclateDistance(this.Pattern, path);
                if (dist < min)
                {
                    min = dist;
                }
            }

            this.distance = min;
            return this.Distance;
        }

        private List<string> GetFullPaths(Document document)
        {
            string path = document.XDocument.Name.LocalName;
            List<string> result = new List<string>();
            GetFullPathsRecursive(ref result, document.XDocument, path);

            return result;
        }

        private void GetFullPathsRecursive(ref List<string> result, XElement element, string path)
        {
            if (!element.HasElements)
            {
                result.Add(path);
            }
            else
            {
                foreach (var e in element.Elements())
                {
                    GetFullPathsRecursive(ref result, e, path + "/" + e.Name.LocalName);
                }
            }
        }
	}
}
