﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Linq;
using XmlClassifier.Algorithms;
using XmlClassifier.Algorithms.DistanceMeasures.PTED;
using XmlClassifier.Helpers;

namespace XmlClassifier.DataStructures.Patterns
{
	[Serializable()]
	public class TreePattern : BasePattern
	{
		public enum Duplicates
		{
			Leave,
			RemoveEqual,
			RemoveEmbedded
		}

		public string Pattern { get; set; }

		private double? size;

		public override double Size
		{
			get
			{
				if (size.HasValue)
				{
					return size.Value;
				}
				else
				{
					return this.XPattern.DescendantNodesAndSelf().Count();
				}
			}
			set
			{
				this.size = value;
			}
		}

		public override List<BasePattern> GetPatterns(List<Document> documents, double minsup)
		{
			List<BasePattern> result = new List<BasePattern>();

			int numberOfClasses = documents.Select(x => x.ExpectedClass).Distinct().Count();
			for (int i = 0; i < numberOfClasses; i++)
			{
				var tmpDocuments = documents.Where(x => x.ExpectedClass == i).ToList();
				var tmp = TreeMiner.ProcessDatabase(tmpDocuments, (int)Math.Round(minsup * (double)tmpDocuments.Count));
				tmp.ForEach(x => { x.Class = i; });

				result.AddRange(tmp);
			}

			TreePattern.CalculateConfidence(documents, result);

			TreePattern.NormalizeWeights(result);

			return result;
		}

		public override bool Matches(Document document)
		{
			//pozycja w dokumencie - tylko rośnie
			int docPos = 0;
			//Miejsce początku wzorca przechowywane na potrzeby ewentualnej kolejnej iteracji,
			//żeby nie trzeba było szukać. Jeśli == -1 to znaczy, że nie znaleziono początku wzorca
			//i należy zacząć poszukiwanie od docPos.
			int nextHop = -1;
			bool patternFound = false;
			bool eof = false;
			string[] doc = document.Content.Split(' ');
			int[,] pat = ParsePattern(this.Pattern.Split(' '));

			while (!patternFound && !eof)
			{
				int docLev = 0;

				if (nextHop >= 0)
				{
					docPos = nextHop;
				}
				nextHop = -1;

				if (docPos >= doc.Length)
				{
					eof = true;
				}

				patternFound = ContainsPattern(ref docPos, ref nextHop, doc, pat, ref docLev);
			}

			if (patternFound)
			{
				return true;
			}

			return false;
		}

		public override int CountMatches(Document document)
		{
			//pozycja w dokumencie - tylko rośnie
			int docPos = 0;
			//Miejsce początku wzorca przechowywane na potrzeby ewentualnej kolejnej iteracji,
			//żeby nie trzeba było szukać. Jeśli == -1 to znaczy, że nie znaleziono początku wzorca
			//i należy zacząć poszukiwanie od docPos.
			int nextHop = -1;
			int patternsFound = 0;
			bool eof = false;
			string[] doc = document.Content.Split(' ');
			int[,] pat = ParsePattern(this.Pattern.Split(' '));

			while (!eof)
			{
				int docLev = 0;

				if (nextHop >= 0)
				{
					docPos = nextHop;
				}
				nextHop = -1;

				if (docPos >= doc.Length)
				{
					eof = true;
				}

				if (ContainsPattern(ref docPos, ref nextHop, doc, pat, ref docLev))
				{
					patternsFound++;
					docPos++;
				}
			}

			return patternsFound;
		}

		public override double CalculateDistance(Document document)
		{
			this.distance = PTED.Dynamic(this.XPattern, document.XDocument) / (double)this.XPattern.DescendantsAndSelf().Count();
			return this.Distance;
		}

		[field: NonSerialized()]
		private XElement xPattern;
		public XElement XPattern
		{
			get
			{
				if (this.xPattern == null)
				{
					this.xPattern = this.Pattern.Parse();
				}

				return this.xPattern;
			}
		}

		private bool ContainsPattern(ref int docPos, ref int nextHop, string[] doc, int[,] pat, ref int docLev)
		{
			for (int i = 0; i < pat.GetLength(0); i++)
			{
				while (docPos < doc.Length && (doc[docPos] != pat[i, 0].ToString() || docLev != pat[i, 1]))
				{
					if (i > 0)
					{
						if (int.Parse(doc[docPos]) < 0)
						{
							docLev--;

							if (IsLevelCrossed(pat, docLev, i))
							{
								for (int patIdx = i; patIdx >= 0; patIdx--)
								{
									//todo sprawdzic
									if (pat[patIdx, 1] == docLev)
									{
										i = patIdx;
										break;
									}
								}
							}
						}
						else
						{
							docLev++;
						}

						//Znaleziono pierwszy element wzorca -> ustawiamy nextHop
						if (doc[docPos] == pat[0, 0].ToString() && nextHop == -1)
						{
							nextHop = docPos;
						}
					}

					docPos++;
				}

				//Znaleziono pierwszy element wzorca (będący równocześnie innym elementem wzorca) -> ustawiamy nextHop
				if (docPos < doc.Length && doc[docPos] == pat[0, 0].ToString() && i > 0 && nextHop == -1)
				{
					nextHop = docPos;
				}

				if (docPos >= doc.Length)
				{
					//dojechaliśmy do końca elementu bez znalezienia kolejnego elementu wzorca
					return false;
				}
				else if (i == pat.GetLength(0) - 1)
				{
					//znaleźliśmy kolejny element wzorca
					//był to ostatni element wzorca
					return true;
				}
				else
				{
					//znaleźliśmy kolejny element wzorca
					//i nie był to ostatni element wzorca
					docLev++;
					docPos++;
				}
			}

			return false;
		}

		private int[,] ParsePattern(string[] patSplit)
		{
			int[,] pat = new int[patSplit.Count(x => int.Parse(x) >= 0), 2];

			int patLev = 0;
			int patPos = 0;
			for (int i = 0; i < patSplit.Length; i++)
			{
				int patVal = int.Parse(patSplit[i]);
				if (patVal < 0)
				{
					patLev--;
				}
				else
				{
					pat[patPos, 0] = patVal;
					pat[patPos, 1] = patLev;
					patLev++;
					patPos++;
				}
			}
			return pat;
		}

		private bool IsLevelCrossed(int[,] pat, int docLev, int i)
		{
			if ((i > 0 && docLev <= pat[i, 1] - 1) || (i == 0 && docLev <= 0))
			{
				//przekroczono poziom
				return true;
			}
			return false;
		}

		public static void RemoveDuplicatePatterns(List<BasePattern> result, Duplicates duplicateTreatment)
		{
			switch (duplicateTreatment)
			{
				case Duplicates.RemoveEqual:
					for (int i = 0; i < result.Count; i++)
					{
						var duplicates = result.FindAll(x => ((TreePattern)x).Pattern == ((TreePattern)result[i]).Pattern);
						if (duplicates.Count > 1)
						{
							foreach (var r in duplicates)
							{
								result.Remove(r);
							}
							i--;
						}
					}
					break;

				case Duplicates.RemoveEmbedded:
					for (int i = 0; i < result.Count; i++)
					{
						var duplicates = result.FindAll(x => PTED.Dynamic(((TreePattern)x).XPattern, ((TreePattern)result[i]).XPattern) == 0 || PTED.Dynamic(((TreePattern)result[i]).XPattern, ((TreePattern)x).XPattern) == 0);
						if (duplicates.Count > 1)
						{
							foreach (var r in duplicates)
							{
								result.Remove(r);
							}
							i--;
						}
					}
					break;

				case Duplicates.Leave:
				default:
					break;
			}
		}

		public static void CalculateConfidence(IEnumerable<Document> documents, IEnumerable<BasePattern> patterns)
		{
			foreach (var p in patterns)
			{
				int globalSupport = 0;
				foreach (var d in documents)
				{
					if (PTED.Dynamic(((TreePattern)p).XPattern, d.XDocument) == 0)// p.Matches(d))
					{
						globalSupport++;
					}
				}
				p.Confidence = p.Support / (double)globalSupport;
			}
		}
	}
}
